from typing import Any

from pydantic import BaseModel, Field


class AuthorizeError(BaseModel):
    error: str = Field(..., regex="invalid_redirect_uri|invalid_client")
    error_description: Any | None = None
    error_uri: str | None = None

    @classmethod
    def get_invalid_redirect_uri(cls, error_description: Any | None = None):
        return cls(error="invalid_redirect_uri", error_description=error_description)

    @classmethod
    def get_invalid_client(cls, error_description: Any | None = None):
        return cls(error="invalid_client", error_description=error_description)


class AuthorizeRedirectError(BaseModel):
    error: str = Field(
        ...,
        regex="invalid_request|unauthorized_client|access_denied|unsupported_response_type|invalid_scope|server_error|temporarily_unavailable|interaction_required|login_required|account_selection_required|consent_required|invalid_request_uri|invalid_request_object|request_not_supported|request_uri_not_supported|registration_not_supported",
    )
    error_description: Any | None = None
    error_uri: str | None = None

    @classmethod
    def get_invalid_request(cls, error_description: Any | None = None):
        return cls(error="invalid_request", error_description=error_description)

    @classmethod
    def get_invalid_scope(cls, error_description: Any | None = None):
        return cls(error="invalid_scope", error_description=error_description)

    @classmethod
    def get_login_required(cls, error_description: Any | None = None):
        return cls(error="login_required", error_description=error_description)

    @classmethod
    def get_consent_required(cls, error_description: Any | None = None):
        return cls(error="consent_required", error_description=error_description)

    @classmethod
    def get_request_not_supported(cls, error_description: Any | None = None):
        return cls(error="request_not_supported", error_description=error_description)


class LoginError(BaseModel):
    error: str = Field(..., regex="invalid_session|registration_disabled")
    error_description: Any | None = None
    error_uri: str | None = None

    @classmethod
    def get_invalid_session(cls, error_description: Any | None = None):
        return cls(error="invalid_session", error_description=error_description)

    @classmethod
    def get_registration_disabled(cls, error_description: Any | None = None):
        return cls(error="registration_disabled", error_description=error_description)


class TokenResponse(BaseModel):
    access_token: str
    id_token: str
    token_type: str = Field("bearer", regex="bearer")
    expires_in: int
    refresh_token: str | None = None


class TokenError(BaseModel):
    error: str = Field(
        ...,
        regex="invalid_request|invalid_client|invalid_grant|unauthorized_client|unsupported_grant_type|invalid_scope",
    )
    error_description: Any | None = None
    error_uri: str | None = None

    @classmethod
    def get_invalid_request(cls):
        return cls(error="invalid_request")

    @classmethod
    def get_invalid_client(cls):
        return cls(error="invalid_client")

    @classmethod
    def get_invalid_grant(cls):
        return cls(error="invalid_grant")

    @classmethod
    def get_unsupported_grant_type(cls):
        return cls(error="unsupported_grant_type")

    @classmethod
    def get_invalid_scope(cls):
        return cls(error="invalid_scope")


class LogoutError(BaseModel):
    error: str = Field(..., regex="invalid_request")
    error_description: Any | None = None
    error_uri: str | None = None

    @classmethod
    def get_invalid_request(cls, error_description: Any | None = None):
        return cls(error="invalid_request", error_description=error_description)



from sqlalchemy import select

from fief.models import OAuthProvider
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class OAuthProviderRepository(
    BaseRepository[OAuthProvider], UUIDRepositoryMixin[OAuthProvider]
):
    model = OAuthProvider

    async def all_by_name_and_provider(self) -> list[OAuthProvider]:
        return await self.list(
            select(OAuthProvider).order_by(OAuthProvider.name, OAuthProvider.provider)
        )



import uuid

import dramatiq

from fief.models import Role, UserPermission
from fief.repositories import RoleRepository, UserPermissionRepository
from fief.tasks.base import ObjectDoesNotExistTaskError, TaskBase


class OnUserRoleCreated(TaskBase):
    __name__ = "on_user_role_created"

    async def run(self, user_id: str, role_id: str, workspace_id: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))

        async with self.get_workspace_session(workspace) as session:
            role_repository = RoleRepository(session)
            user_permission_repository = UserPermissionRepository(session)

            role = await role_repository.get_by_id(uuid.UUID(role_id))

            if role is None:
                raise ObjectDoesNotExistTaskError(Role, role_id)

            user_permissions: list[UserPermission] = []
            for permission in role.permissions:
                user_permissions.append(
                    UserPermission(
                        user_id=uuid.UUID(user_id),
                        permission_id=permission.id,
                        from_role_id=role.id,
                    )
                )
            await user_permission_repository.create_many(user_permissions)


class OnUserRoleDeleted(TaskBase):
    __name__ = "on_user_role_deleted"

    async def run(self, user_id: str, role_id: str, workspace_id: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))

        async with self.get_workspace_session(workspace) as session:
            user_permission_repository = UserPermissionRepository(session)
            await user_permission_repository.delete_by_user_and_role(
                uuid.UUID(user_id), uuid.UUID(role_id)
            )


on_user_role_created = dramatiq.actor(OnUserRoleCreated())
on_user_role_deleted = dramatiq.actor(OnUserRoleDeleted())



import ipaddress
import re

LOCALHOST_HOST_PATTERN = re.compile(r"([^\.]+\.)?localhost(\d+)?", flags=re.IGNORECASE)


def is_localhost(host: str) -> bool:
    try:
        return ipaddress.IPv4Address(host).is_private
    except ValueError:
        return LOCALHOST_HOST_PATTERN.match(host) is not None


__all__ = ["is_localhost"]



from typing import Any, TypeVar

from fastapi import Request, Response

from fief.dependencies.users import UserManager
from fief.models import (
    OAuthAccount,
    RegistrationSession,
    RegistrationSessionFlow,
    Tenant,
    User,
    UserField,
)
from fief.repositories import OAuthAccountRepository, RegistrationSessionRepository
from fief.schemas.user import UF, UserCreateInternal
from fief.settings import settings

ResponseType = TypeVar("ResponseType", bound=Response)


class RegistrationFlow:
    def __init__(
        self,
        registration_session_repository: RegistrationSessionRepository,
        oauth_account_repository: OAuthAccountRepository,
        user_manager: UserManager,
        user_create_internal_model: type[UserCreateInternal[UF]],
        user_fields: list[UserField],
    ) -> None:
        self.registration_session_repository = registration_session_repository
        self.oauth_account_repository = oauth_account_repository
        self.user_manager = user_manager
        self.user_create_internal_model = user_create_internal_model
        self.user_fields = user_fields

    async def create_registration_session(
        self,
        response: ResponseType,
        flow: RegistrationSessionFlow,
        *,
        tenant: Tenant,
        oauth_account: OAuthAccount | None = None,
    ) -> ResponseType:
        email = None
        oauth_account_id = None
        if oauth_account is not None:
            email = oauth_account.account_email
            oauth_account_id = oauth_account.id

        registration_session = RegistrationSession(
            flow=flow,
            email=email,
            oauth_account_id=oauth_account_id,
            tenant_id=tenant.id,
        )
        registration_session = await self.registration_session_repository.create(
            registration_session
        )

        response.set_cookie(
            settings.registration_session_cookie_name,
            registration_session.token,
            domain=settings.registration_session_cookie_domain,
            secure=settings.registration_session_cookie_secure,
        )

        return response

    async def delete_registration_session(
        self, response: ResponseType, registration_session: RegistrationSession
    ) -> ResponseType:
        response.delete_cookie(
            settings.registration_session_cookie_name,
            domain=settings.registration_session_cookie_domain,
        )
        await self.registration_session_repository.delete(registration_session)
        return response

    async def create_user(
        self,
        data: dict[str, Any],
        tenant: Tenant,
        registration_session: RegistrationSession,
        request: Request | None = None,
    ) -> User:
        user_create_dict: dict[str, Any] = {**data, "tenant_id": tenant.id}
        if "password" not in user_create_dict:
            user_create_dict["password"] = self.user_manager.password_helper.generate()
        user_create = self.user_create_internal_model(**user_create_dict)
        user = await self.user_manager.create_with_fields(
            user_create,
            user_fields=self.user_fields,
            safe=True,
            request=request,
        )

        if registration_session.oauth_account is not None:
            registration_session.oauth_account.user = user
            await self.oauth_account_repository.update(
                registration_session.oauth_account
            )

        return user



from fastapi import Depends, HTTPException, Request, status
from fastapi.security import APIKeyCookie
from fief_client import FiefUserInfo

from fief.crypto.token import get_token_hash
from fief.dependencies.main_repositories import get_main_repository
from fief.models import AdminSessionToken
from fief.repositories import AdminSessionTokenRepository
from fief.settings import settings

cookie_scheme = APIKeyCookie(
    name=settings.fief_admin_session_cookie_name, auto_error=False
)


async def get_optional_admin_session_token(
    token: str | None = Depends(cookie_scheme),
    repository: AdminSessionTokenRepository = Depends(
        get_main_repository(AdminSessionTokenRepository)
    ),
) -> AdminSessionToken | None:
    if token is None:
        return None
    token_hash = get_token_hash(token)
    session_token = await repository.get_by_token(token_hash)
    return session_token


async def get_admin_session_token(
    request: Request,
    admin_session_token: AdminSessionToken
    | None = Depends(get_optional_admin_session_token),
) -> AdminSessionToken:
    if admin_session_token is None:
        raise HTTPException(
            status_code=status.HTTP_307_TEMPORARY_REDIRECT,
            headers={"Location": request.url_for("dashboard.auth:login")},
        )
    return admin_session_token


async def get_userinfo(
    session_token: AdminSessionToken = Depends(get_admin_session_token),
) -> FiefUserInfo:
    return session_token.userinfo



from fastapi import APIRouter, Depends, Header, Request, status

from fief import schemas
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.permission import PermissionCreateForm
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.permission import (
    get_paginated_permissions,
    get_permission_by_id_or_404,
)
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Permission
from fief.repositories import PermissionRepository
from fief.services.webhooks.models import (
    PermissionCreated,
    PermissionDeleted,
)
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Name", "name", "name_column", ordering="name"),
        DatatableColumn("Codename", "codename", "codename_column", ordering="codename"),
        DatatableColumn("Actions", "actions", "actions_column"),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["name", "codename", "actions"])
    ),
    paginated_permissions: PaginatedObjects[Permission] = Depends(
        get_paginated_permissions
    ),
):
    permissions, count = paginated_permissions
    return {
        "permissions": permissions,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


async def get_list_template(hx_combobox: bool = Header(False)) -> str:
    if hx_combobox:
        return "admin/permissions/list_combobox.html"
    return "admin/permissions/list.html"


async def get_form_helper(
    request: Request,
    template: str = Depends(get_list_template),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
) -> FormHelper[PermissionCreateForm]:
    form_helper = FormHelper(
        PermissionCreateForm,
        template,
        request=request,
        context={**context, **list_context},
    )
    await form_helper.get_form()
    return form_helper


@router.api_route("/", methods=["GET", "POST"], name="dashboard.permissions:list")
async def list_permissions(
    request: Request,
    form_helper: FormHelper[PermissionCreateForm] = Depends(get_form_helper),
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        existing_permission = await repository.get_by_codename(form.data["codename"])
        if existing_permission is not None:
            form.codename.errors.append(
                "A permission already exists with this codename. A codename must be unique within your workspace."
            )
            return await form_helper.get_error_response(
                "A permission already exists with this codename. A codename must be unique within your workspace.",
                "permission_codename_already_exists",
            )

        permission = Permission()
        form.populate_obj(permission)
        permission = await repository.create(permission)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, permission)
        trigger_webhooks(PermissionCreated, permission, schemas.permission.Permission)

        return HXRedirectResponse(
            request.url_for("dashboard.permissions:list"),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(permission.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/delete", methods=["GET", "DELETE"], name="dashboard.permissions:delete"
)
async def delete_permission(
    request: Request,
    form_helper: FormHelper[PermissionCreateForm] = Depends(get_form_helper),
    permission: Permission = Depends(get_permission_by_id_or_404),
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if request.method == "DELETE":
        await repository.delete(permission)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, permission)
        trigger_webhooks(PermissionDeleted, permission, schemas.permission.Permission)

        return HXRedirectResponse(
            request.url_for("dashboard.permissions:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/permissions/delete.html",
            {**form_helper.context, "permission": permission},
        )



from pydantic import BaseModel

from fief.schemas.generics import CreatedUpdatedAt, UUIDSchema


class PermissionCreate(BaseModel):
    name: str
    codename: str


class PermissionUpdate(BaseModel):
    name: str | None
    codename: str | None


class BasePermission(UUIDSchema, CreatedUpdatedAt):
    name: str
    codename: str


class Permission(BasePermission):
    pass


class PermissionEmbedded(BasePermission):
    pass



"""Add expires_at to LoginSession and AuthorizationCode

Revision ID: 2b8bf1c3378e
Revises: f3784e0c8524
Create Date: 2022-07-06 10:44:30.600940

"""

import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "2b8bf1c3378e"
down_revision = "f3784e0c8524"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("DELETE FROM fief_authorization_codes")
    op.execute("DELETE FROM fief_login_sessions")

    connection = op.get_bind()

    server_default = None
    if connection.dialect.name == "sqlite":
        server_default = "0"

    op.add_column(
        "fief_authorization_codes",
        sa.Column(
            "expires_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=server_default,
            nullable=False,
        ),
    )
    op.create_index(
        op.f("ix_fief_authorization_codes_expires_at"),
        "fief_authorization_codes",
        ["expires_at"],
        unique=False,
    )
    op.add_column(
        "fief_login_sessions",
        sa.Column(
            "expires_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=server_default,
            nullable=False,
        ),
    )
    op.create_index(
        op.f("ix_fief_login_sessions_expires_at"),
        "fief_login_sessions",
        ["expires_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_fief_login_sessions_expires_at"), table_name="fief_login_sessions"
    )
    op.drop_column("fief_login_sessions", "expires_at")
    op.drop_index(
        op.f("ix_fief_authorization_codes_expires_at"),
        table_name="fief_authorization_codes",
    )
    op.drop_column("fief_authorization_codes", "expires_at")
    # ### end Alembic commands ###



from pydantic import UUID4
from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column

from fief.models.base import MainBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.workspace import Workspace


class AdminAPIKey(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "admin_api_key"

    name: Mapped[str] = mapped_column(String(length=255), nullable=False)
    token: Mapped[str] = mapped_column(String(length=255), unique=True, nullable=False)
    workspace_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Workspace.id, ondelete="CASCADE"), nullable=False
    )

    def __repr__(self) -> str:
        return f"AdminAPIKey(id={self.id})"



from pydantic import UUID4

from fief.schemas.generics import CreatedUpdatedAt, UUIDSchema
from fief.schemas.webhook import WebhookEventType


class BaseWebhookLog(UUIDSchema, CreatedUpdatedAt):
    webhook_id: UUID4
    event: WebhookEventType
    attempt: int
    payload: str
    success: bool
    response: str | None
    error_type: str | None
    error_message: str | None


class WebhookLog(BaseWebhookLog):
    pass



from wtforms import Form, IntegerField, SelectField, StringField, URLField, validators

from fief.forms import CSRFBaseForm


class ThemeCreateForm(CSRFBaseForm):
    name = StringField("Name", validators=[validators.InputRequired()])


class ThemePagePreviewForm(Form):
    page = SelectField(
        choices=[
            ("login", "Sign in"),
            ("register", "Sign up"),
            ("forgot_password", "Forgot password"),
            ("reset_password", "Reset password"),
        ]
    )


class ThemeUpdateForm(CSRFBaseForm):
    name = StringField("Theme's name", validators=[validators.InputRequired()])

    primary_color = StringField(
        "Primary color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )
    primary_color_hover = StringField(
        "Hover primary color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )
    primary_color_light = StringField(
        "Light primary color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )

    input_color = StringField(
        "Form input text color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )
    input_color_background = StringField(
        "Form input background color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )

    light_color = StringField(
        "Light color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )
    light_color_hover = StringField(
        "Hover light color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )

    text_color = StringField(
        "Text color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )
    accent_color = StringField(
        "Accent text color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )

    background_color = StringField(
        "Background color",
        render_kw={"type": "color"},
        validators=[validators.InputRequired()],
    )

    font_size = IntegerField("Base font size", validators=[validators.InputRequired()])
    font_family = StringField("Font family", validators=[validators.InputRequired()])
    font_css_url = URLField(
        "CSS font URL",
        validators=[validators.Optional(), validators.URL(require_tld=False)],
    )



import uuid

import dramatiq

from fief import schemas
from fief.models import UserRole
from fief.repositories import RoleRepository, UserRoleRepository
from fief.services.email_template.contexts import WelcomeContext
from fief.services.email_template.types import EmailTemplateType
from fief.tasks.base import TaskBase, send_task
from fief.tasks.user_permissions import on_user_role_created


class OnAfterRegisterTask(TaskBase):
    __name__ = "on_after_register"

    async def run(self, user_id: str, workspace_id: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))
        user = await self._get_user(uuid.UUID(user_id), workspace)
        tenant = await self._get_tenant(user.tenant_id, workspace)

        # Grant default roles
        async with self.get_workspace_session(workspace) as session:
            role_repository = RoleRepository(session)
            user_role_repository = UserRoleRepository(session)

            default_roles = await role_repository.get_granted_by_default()
            user_roles: list[UserRole] = []
            for role in default_roles:
                existing_user_role = await user_role_repository.get_by_role_and_user(
                    user.id, role.id
                )
                if existing_user_role is None:
                    user_roles.append(UserRole(user_id=user.id, role_id=role.id))
            user_roles = await user_role_repository.create_many(user_roles)
            for user_role in user_roles:
                send_task(
                    on_user_role_created,
                    str(user.id),
                    str(user_role.role_id),
                    str(workspace.id),
                )

        # Send welcome email
        context = WelcomeContext(
            tenant=schemas.tenant.Tenant.from_orm(tenant),
            user=schemas.user.UserEmailContext.from_orm(user),
        )
        async with self._get_email_subject_renderer(
            workspace
        ) as email_subject_renderer:
            subject = await email_subject_renderer.render(
                EmailTemplateType.WELCOME, context
            )

        async with self._get_email_template_renderer(
            workspace
        ) as email_template_renderer:
            html = await email_template_renderer.render(
                EmailTemplateType.WELCOME, context
            )

        self.email_provider.send_email(
            sender=("contact@fief.dev", None),
            recipient=(user.email, None),
            subject=subject,
            html=html,
        )


on_after_register = dramatiq.actor(OnAfterRegisterTask())



from pydantic import BaseModel, root_validator, validator

from fief.db.types import SSL_MODES, DatabaseType
from fief.errors import APIErrorCode


def validate_all_database_settings(cls, values):
    database_type = values.get("database_type")
    required_database_settings = [
        values.get("database_host"),
        values.get("database_port"),
        values.get("database_username"),
        values.get("database_password"),
        values.get("database_name"),
    ]

    if database_type is None and not any(required_database_settings):
        return values

    if database_type is None and any(required_database_settings):
        raise ValueError(APIErrorCode.WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS.value)

    database_name = values.get("database_name")
    if database_type == DatabaseType.SQLITE:
        if database_name is None:
            raise ValueError(
                APIErrorCode.WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS.value
            )
    else:
        if not all(required_database_settings):
            raise ValueError(
                APIErrorCode.WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS.value
            )

    return values


def validate_ssl_mode(ssl_mode, values):
    type = values.get("database_type")

    if type is not None and ssl_mode is not None:
        ssl_mode_type = SSL_MODES[type]
        try:
            ssl_mode_type(ssl_mode)
        except ValueError as e:
            raise ValueError(
                APIErrorCode.WORKSPACE_CREATE_INVALID_SSL_MODE.value
            ) from e

    return ssl_mode


class WorkspaceCreate(BaseModel):
    name: str
    database_type: DatabaseType | None = None
    database_host: str | None = None
    database_port: int | None = None
    database_username: str | None = None
    database_password: str | None = None
    database_name: str | None = None
    database_ssl_mode: str | None = None

    _validate_all_database_settings = root_validator(allow_reuse=True)(
        validate_all_database_settings
    )
    _validate_ssl_mode = validator("database_ssl_mode", allow_reuse=True)(
        validate_ssl_mode
    )



import uuid

import dramatiq
from sqlalchemy.exc import DBAPIError

from fief.logger import logger
from fief.repositories import UserRepository, WorkspaceRepository
from fief.tasks.base import TaskBase


class CountUsersWorkspaceTask(TaskBase):
    __name__ = "count_users_workspace"

    async def run(self, workspace_id: str):
        async with self.get_main_session() as session:
            workspace_repository = WorkspaceRepository(session)
            workspace = await self._get_workspace(uuid.UUID(workspace_id))
            try:
                async with self.get_workspace_session(workspace) as workspace_session:
                    user_repository = UserRepository(workspace_session)
                    workspace.users_count = await user_repository.count_all()
                    await workspace_repository.update(workspace)
            except ConnectionError:
                logger.warning(
                    "Could not connect to workspace", workspace_id=str(workspace.id)
                )
            except DBAPIError as e:
                logger.warning(
                    "An error occured while querying workspace",
                    workspace_id=str(workspace.id),
                    message=str(e),
                )


count_users_workspace = dramatiq.actor(CountUsersWorkspaceTask(), max_retries=0)


class CountUsersTask(TaskBase):
    __name__ = "count_users"

    async def run(self):
        async with self.get_main_session() as session:
            workspace_repository = WorkspaceRepository(session)
            workspaces = await workspace_repository.all()
            for workspace in workspaces:
                self.send_task(count_users_workspace, workspace_id=str(workspace.id))


count_users = dramatiq.actor(CountUsersTask())



"""Add Theme model

Revision ID: f72167c6b223
Revises: ab9cf21c76dd
Create Date: 2023-01-04 10:42:08.479094

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "f72167c6b223"
down_revision = "ab9cf21c76dd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_themes",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("default", sa.Boolean(), nullable=False),
        sa.Column("primary_color", sa.String(length=255), nullable=False),
        sa.Column("primary_color_hover", sa.String(length=255), nullable=False),
        sa.Column("primary_color_light", sa.String(length=255), nullable=False),
        sa.Column("input_color", sa.String(length=255), nullable=False),
        sa.Column("input_color_background", sa.String(length=255), nullable=False),
        sa.Column("light_color", sa.String(length=255), nullable=False),
        sa.Column("light_color_hover", sa.String(length=255), nullable=False),
        sa.Column("text_color", sa.String(length=255), nullable=False),
        sa.Column("accent_color", sa.String(length=255), nullable=False),
        sa.Column("background_color", sa.String(length=255), nullable=False),
        sa.Column("font_size", sa.Integer(), nullable=False),
        sa.Column("font_family", sa.String(length=255), nullable=False),
        sa.Column("font_css_url", sa.String(length=512), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_themes_created_at"), "fief_themes", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_fief_themes_updated_at"), "fief_themes", ["updated_at"], unique=False
    )

    connection = op.get_bind()
    if connection.dialect.name == "sqlite":
        with op.batch_alter_table("fief_tenants") as batch_op:
            batch_op.add_column(
                sa.Column("theme_id", fief.models.generics.GUID(), nullable=True),
            )
            batch_op.create_foreign_key(
                "fief_tenants_theme_id_fkey",
                referent_table="fief_themes",
                local_cols=["theme_id"],
                remote_cols=["id"],
                ondelete="SET NULL",
            )
    else:
        op.add_column(
            "fief_tenants",
            sa.Column("theme_id", fief.models.generics.GUID(), nullable=True),
        )
        op.create_foreign_key(
            "fief_tenants_theme_id_fkey",
            "fief_tenants",
            "fief_themes",
            ["theme_id"],
            ["id"],
            ondelete="SET NULL",
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "fief_tenants", type_="foreignkey")
    op.drop_column("fief_tenants", "theme_id")
    op.drop_index(op.f("ix_fief_themes_updated_at"), table_name="fief_themes")
    op.drop_index(op.f("ix_fief_themes_created_at"), table_name="fief_themes")
    op.drop_table("fief_themes")
    # ### end Alembic commands ###



from pydantic import UUID4

from fief.schemas.generics import BaseModel, CreatedUpdatedAt, UUIDSchema
from fief.schemas.permission import PermissionEmbedded


class RoleCreate(BaseModel):
    name: str
    granted_by_default: bool
    permissions: list[UUID4]


class RoleUpdate(BaseModel):
    name: str | None
    granted_by_default: bool | None
    permissions: list[UUID4] | None


class BaseRole(UUIDSchema, CreatedUpdatedAt):
    name: str
    granted_by_default: bool


class Role(BaseRole):
    permissions: list[PermissionEmbedded]


class RoleEmbedded(BaseRole):
    pass



"""Remove index on AuditLog.message

Revision ID: aa662768d69b
Revises: 2501f74d0c55
Create Date: 2022-11-29 09:31:08.396676

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy import exc

import fief

# revision identifiers, used by Alembic.
revision = "aa662768d69b"
down_revision = "2501f74d0c55"
branch_labels = None
depends_on = None


def upgrade():
    """
    On older versions of Fief, we created an index on AuditLog.message.
    It provoked the bug fief-dev/fief#115 on MySQL.

    We removed its creation in the existing migration, but we still need
    to remove it from existing DB, if it exists.
    """
    bind = op.get_context().bind
    inspector = sa.inspect(bind)
    indexes = inspector.get_indexes("fief_audit_logs")
    for index in indexes:
        if index["name"] == "ix_fief_audit_logs_message":
            op.drop_index("ix_fief_audit_logs_message", table_name="fief_audit_logs")
            break


def downgrade():
    pass



from fastapi import Depends, HTTPException, status
from pydantic import UUID4
from sqlalchemy import select

from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import EmailTemplate
from fief.repositories import EmailTemplateRepository


async def get_paginated_email_templates(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: EmailTemplateRepository = Depends(
        get_workspace_repository(EmailTemplateRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[EmailTemplate] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[EmailTemplate]:
    statement = select(EmailTemplate)
    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_email_template_by_id_or_404(
    id: UUID4,
    repository: EmailTemplateRepository = Depends(
        get_workspace_repository(EmailTemplateRepository)
    ),
) -> EmailTemplate:
    email_template = await repository.get_by_id(id)

    if email_template is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return email_template



from fastapi import APIRouter, Depends, Header, Request, status

from fief import schemas
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.tenant import TenantCreateForm, TenantUpdateForm
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.tenant import get_paginated_tenants, get_tenant_by_id_or_404
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Client, OAuthProvider, Tenant
from fief.repositories import (
    ClientRepository,
    OAuthProviderRepository,
    TenantRepository,
    ThemeRepository,
    UserRepository,
)
from fief.services.webhooks.models import (
    ClientCreated,
    TenantCreated,
    TenantDeleted,
    TenantUpdated,
)
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Name", "name", "name_column", ordering="name"),
        DatatableColumn("Base URL", "base_url", "base_url_column"),
        DatatableColumn(
            "Registration allowed",
            "registration_allowed",
            "registration_allowed_column",
            ordering="registration_allowed",
        ),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["name", "base_url", "registration_allowed"])
    ),
    paginated_tenants: PaginatedObjects[Tenant] = Depends(get_paginated_tenants),
):
    tenants, count = paginated_tenants
    return {
        "tenants": tenants,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


async def get_list_template(hx_combobox: bool = Header(False)) -> str:
    if hx_combobox:
        return "admin/tenants/list_combobox.html"
    return "admin/tenants/list.html"


@router.get("/", name="dashboard.tenants:list")
async def list_tenants(
    template: str = Depends(get_list_template),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(template, {**context, **list_context})


@router.get("/{id:uuid}", name="dashboard.tenants:get")
async def get_tenant(
    tenant: Tenant = Depends(get_tenant_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/tenants/get.html",
        {**context, **list_context, "tenant": tenant},
    )


@router.api_route("/create", methods=["GET", "POST"], name="dashboard.tenants:create")
async def create_tenant(
    request: Request,
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
    client_repository: ClientRepository = Depends(
        get_workspace_repository(ClientRepository)
    ),
    theme_repository: ThemeRepository = Depends(
        get_workspace_repository(ThemeRepository)
    ),
    oauth_provider_repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        TenantCreateForm,
        "admin/tenants/create.html",
        request=request,
        context={**context, **list_context},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        tenant = Tenant()

        theme_id = form.data["theme"]
        if theme_id is not None:
            theme = await theme_repository.get_by_id(theme_id)
            if theme is None:
                form.theme.errors.append("Unknown theme.")
                return await form_helper.get_error_response(
                    "Unknown theme.", "unknown_theme"
                )
            form.theme.data = theme

        oauth_providers_ids = form.data["oauth_providers"]
        oauth_providers: list[OAuthProvider] = []
        for oauth_provider_id in oauth_providers_ids:
            oauth_provider = await oauth_provider_repository.get_by_id(
                oauth_provider_id
            )
            if oauth_provider is None:
                form.oauth_providers.errors.append("Unknown OAuth Provider.")
                return await form_helper.get_error_response(
                    "Unknown OAuth Provider.", "unknown_oauth_provider"
                )
            oauth_providers.append(oauth_provider)
            form.oauth_providers.data = oauth_providers

        form.populate_obj(tenant)
        tenant.slug = await repository.get_available_slug(tenant.name)
        tenant = await repository.create(tenant)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, tenant)
        trigger_webhooks(TenantCreated, tenant, schemas.tenant.Tenant)

        client = Client(
            name=f"{tenant.name}'s client",
            first_party=True,
            tenant=tenant,
            redirect_uris=["http://localhost:8000/docs/oauth2-redirect"],
        )
        await client_repository.create(client)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, client)
        trigger_webhooks(ClientCreated, client, schemas.client.Client)

        return HXRedirectResponse(
            request.url_for("dashboard.tenants:get", id=tenant.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(tenant.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.tenants:update",
)
async def update_tenant(
    request: Request,
    tenant: Tenant = Depends(get_tenant_by_id_or_404),
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
    theme_repository: ThemeRepository = Depends(
        get_workspace_repository(ThemeRepository)
    ),
    oauth_provider_repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        TenantUpdateForm,
        "admin/tenants/edit.html",
        object=tenant,
        request=request,
        context={**context, **list_context, "tenant": tenant},
    )
    form = await form_helper.get_form()
    form.oauth_providers.choices = [
        (oauth_provider.id, oauth_provider.display_name)
        for oauth_provider in tenant.oauth_providers
    ]

    if await form_helper.is_submitted_and_valid():
        theme_id = form.data["theme"]
        if theme_id is not None:
            theme = await theme_repository.get_by_id(theme_id)
            if theme is None:
                form.theme.errors.append("Unknown theme.")
                return await form_helper.get_error_response(
                    "Unknown theme.", "unknown_theme"
                )
            form.theme.data = theme

        tenant.oauth_providers = []
        for oauth_provider_id in form.data["oauth_providers"]:
            oauth_provider = await oauth_provider_repository.get_by_id(
                oauth_provider_id
            )
            if oauth_provider is None:
                form.oauth_providers.errors.append("Unknown OAuth Provider.")
                return await form_helper.get_error_response(
                    "Unknown OAuth Provider.", "unknown_oauth_provider"
                )
            tenant.oauth_providers.append(oauth_provider)

        del form.oauth_providers
        form.populate_obj(tenant)

        await repository.update(tenant)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, tenant)
        trigger_webhooks(TenantUpdated, tenant, schemas.tenant.Tenant)

        return HXRedirectResponse(
            request.url_for("dashboard.tenants:get", id=tenant.id)
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/delete",
    methods=["GET", "DELETE"],
    name="dashboard.tenants:delete",
)
async def delete_tenant(
    request: Request,
    tenant: Tenant = Depends(get_tenant_by_id_or_404),
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
    user_repository: UserRepository = Depends(get_workspace_repository(UserRepository)),
    client_repository: ClientRepository = Depends(
        get_workspace_repository(ClientRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if request.method == "DELETE":
        await repository.delete(tenant)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, tenant)
        trigger_webhooks(TenantDeleted, tenant, schemas.tenant.Tenant)

        return HXRedirectResponse(
            request.url_for("dashboard.tenants:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )

    users_count = await user_repository.count_by_tenant(tenant.id)
    clients_count = await client_repository.count_by_tenant(tenant.id)
    return templates.TemplateResponse(
        "admin/tenants/delete.html",
        {
            **context,
            **list_context,
            "tenant": tenant,
            "users_count": users_count,
            "clients_count": clients_count,
        },
    )



from wtforms import (
    BooleanField,
    FieldList,
    IntegerField,
    SelectField,
    StringField,
    validators,
)

from fief.apps.dashboard.validators import RedirectURLValidator
from fief.forms import ComboboxSelectField, CSRFBaseForm
from fief.models import ClientType


class BaseClientForm(CSRFBaseForm):
    name = StringField("Name", validators=[validators.InputRequired()])
    first_party = BooleanField("First-party")
    client_type = SelectField(
        "Type",
        choices=ClientType.choices(),
        coerce=ClientType.coerce,
        default=ClientType.CONFIDENTIAL.value,
        validators=[validators.InputRequired()],
    )
    redirect_uris = FieldList(
        StringField(validators=[validators.InputRequired(), RedirectURLValidator()]),
        label="Redirect URIs",
        min_entries=1,
    )


class ClientCreateForm(BaseClientForm):
    tenant = ComboboxSelectField(
        "Tenant",
        query_endpoint_path="/admin/tenants/",
        validators=[validators.InputRequired(), validators.UUID()],
    )


class ClientUpdateForm(BaseClientForm):
    pass


class ClientLifetimesForm(CSRFBaseForm):
    authorization_code_lifetime_seconds = IntegerField(
        "Authorization code",
        description="In seconds",
        validators=[validators.InputRequired(), validators.NumberRange(min=0)],
    )
    access_id_token_lifetime_seconds = IntegerField(
        "Access and ID Token",
        description="In seconds",
        validators=[validators.InputRequired(), validators.NumberRange(min=0)],
    )
    refresh_token_lifetime_seconds = IntegerField(
        "Refresh Token",
        description="In seconds",
        validators=[validators.InputRequired(), validators.NumberRange(min=0)],
    )



import functools
import uuid
from datetime import datetime, timedelta, timezone
from typing import TypeVar

from pydantic import UUID4
from sqlalchemy import TIMESTAMP
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Mapped, MappedColumn, mapped_column
from sqlalchemy.sql import func
from sqlalchemy.types import CHAR, TypeDecorator


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(36), storing as regular strings.
    """

    class UUIDChar(CHAR):
        python_type = UUID4  # type: ignore

    impl = UUIDChar
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(36))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return str(uuid.UUID(value))
            else:
                return str(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


class BaseModel:
    pass


class UUIDModel(BaseModel):
    id: Mapped[UUID4] = mapped_column(GUID, primary_key=True, default=uuid.uuid4)


def now_utc():
    return datetime.now(timezone.utc)


class TIMESTAMPAware(TypeDecorator):
    """
    MySQL and SQLite will always return naive-Python datetimes.

    We store everything as UTC, but we want to have
    only offset-aware Python datetimes, even with MySQL and SQLite.
    """

    impl = TIMESTAMP
    cache_ok = True

    def process_result_value(self, value: datetime | None, dialect):
        if value is not None and dialect.name != "postgresql":
            return value.replace(tzinfo=timezone.utc)
        return value


class CreatedUpdatedAt(BaseModel):
    created_at: Mapped[datetime] = mapped_column(
        TIMESTAMPAware(timezone=True),
        nullable=False,
        index=True,
        default=now_utc,
        server_default=func.now(),
    )
    updated_at: Mapped[datetime] = mapped_column(
        TIMESTAMPAware(timezone=True),
        nullable=False,
        index=True,
        default=now_utc,
        server_default=func.now(),
        onupdate=now_utc,
    )


def _get_default_expires_at(timedelta_seconds: int) -> datetime:
    return datetime.now(timezone.utc) + timedelta(seconds=timedelta_seconds)


class ExpiresAt(BaseModel):
    @declared_attr
    def expires_at(cls) -> MappedColumn[TIMESTAMPAware]:
        try:
            default_lifetime_seconds = getattr(
                cls,
                "__lifetime_seconds__",
            )
            default = functools.partial(
                _get_default_expires_at, timedelta_seconds=default_lifetime_seconds
            )
        except AttributeError:
            default = None
        return mapped_column(
            TIMESTAMPAware(timezone=True), nullable=False, index=True, default=default
        )


M = TypeVar("M", bound=BaseModel)
M_UUID = TypeVar("M_UUID", bound=UUIDModel)
M_EXPIRES_AT = TypeVar("M_EXPIRES_AT", bound=ExpiresAt)



from fastapi import APIRouter, Depends, Response, status

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.webhook import (
    get_paginated_webhook_logs,
    get_paginated_webhooks,
    get_webhook_by_id_or_404,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Webhook, WebhookLog
from fief.repositories import WebhookRepository
from fief.schemas.generics import PaginatedResults

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/", name="webhooks:list", response_model=PaginatedResults[schemas.webhook.Webhook]
)
async def list_webhooks(
    paginated_webhooks: PaginatedObjects[Webhook] = Depends(get_paginated_webhooks),
) -> PaginatedResults[schemas.webhook.Webhook]:
    webhooks, count = paginated_webhooks
    return PaginatedResults(
        count=count,
        results=[schemas.webhook.Webhook.from_orm(webhook) for webhook in webhooks],
    )


@router.post(
    "/",
    name="webhooks:create",
    response_model=schemas.webhook.WebhookSecret,
    status_code=status.HTTP_201_CREATED,
)
async def create_webhook(
    webhook_create: schemas.webhook.WebhookCreate,
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
) -> schemas.webhook.WebhookSecret:
    webhook = Webhook(**webhook_create.dict())
    webhook = await repository.create(webhook)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, webhook)

    return schemas.webhook.WebhookSecret.from_orm(webhook)


@router.get("/{id:uuid}", name="webhooks:get", response_model=schemas.webhook.Webhook)
async def get_webhook(webhook: Webhook = Depends(get_webhook_by_id_or_404)) -> Webhook:
    return webhook


@router.patch(
    "/{id:uuid}", name="webhooks:update", response_model=schemas.webhook.Webhook
)
async def update_webhook(
    webhook_update: schemas.webhook.WebhookUpdate,
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
) -> schemas.webhook.Webhook:
    webhook_update_dict = webhook_update.dict(exclude_unset=True)
    for field, value in webhook_update_dict.items():
        setattr(webhook, field, value)

    await repository.update(webhook)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, webhook)

    return schemas.webhook.Webhook.from_orm(webhook)


@router.post(
    "/{id:uuid}/secret",
    name="webhooks:secret",
    response_model=schemas.webhook.WebhookSecret,
)
async def regenerate_webhook_secret(
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    webhook.regenerate_secret()
    await repository.update(webhook)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, webhook)

    return webhook


@router.delete(
    "/{id:uuid}",
    name="webhooks:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_role(
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    await repository.delete(webhook)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, webhook)


@router.get(
    "/{id:uuid}/logs",
    name="webhooks:logs",
    response_model=PaginatedResults[schemas.webhook_log.WebhookLog],
)
async def list_webhook_logs(
    paginated_webhook_logs: PaginatedObjects[WebhookLog] = Depends(
        get_paginated_webhook_logs
    ),
):
    webhook_logs, count = paginated_webhook_logs
    return PaginatedResults(
        count=count,
        results=[
            schemas.webhook_log.WebhookLog.from_orm(webhook_log)
            for webhook_log in webhook_logs
        ],
    )



from pydantic import UUID4, AnyUrl, Field, SecretStr, validator

from fief.errors import APIErrorCode
from fief.models.client import ClientType
from fief.schemas.generics import BaseModel, CreatedUpdatedAt, UUIDSchema
from fief.schemas.tenant import TenantEmbedded
from fief.services.localhost import is_localhost


def validate_redirect_uri(url: AnyUrl) -> AnyUrl:
    if url.scheme == "http" and (url.host is None or not is_localhost(url.host)):
        raise ValueError(APIErrorCode.CLIENT_HTTPS_REQUIRED_ON_REDIRECT_URIS.value)
    return url


class ClientCreate(BaseModel):
    name: str
    first_party: bool
    client_type: ClientType
    redirect_uris: list[AnyUrl] = Field(..., min_items=1)
    authorization_code_lifetime_seconds: int | None = Field(None, ge=0)
    access_id_token_lifetime_seconds: int | None = Field(None, ge=0)
    refresh_token_lifetime_seconds: int | None = Field(None, ge=0)
    tenant_id: UUID4

    _validate_redirect_uri = validator(
        "redirect_uris", each_item=True, allow_reuse=True
    )(validate_redirect_uri)


class ClientUpdate(BaseModel):
    name: str | None
    first_party: bool | None
    client_type: ClientType | None
    redirect_uris: list[AnyUrl] | None = Field(None, min_items=1)
    authorization_code_lifetime_seconds: int | None = Field(None, ge=0)
    access_id_token_lifetime_seconds: int | None = Field(None, ge=0)
    refresh_token_lifetime_seconds: int | None = Field(None, ge=0)

    _validate_redirect_uri = validator(
        "redirect_uris", each_item=True, allow_reuse=True
    )(validate_redirect_uri)


class BaseClient(UUIDSchema, CreatedUpdatedAt):
    name: str
    first_party: bool
    client_type: ClientType
    client_id: str
    client_secret: str
    redirect_uris: list[AnyUrl]
    authorization_code_lifetime_seconds: int
    access_id_token_lifetime_seconds: int
    refresh_token_lifetime_seconds: int
    tenant_id: UUID4


class Client(BaseClient):
    tenant: TenantEmbedded
    encrypt_jwk: SecretStr | None



from pydantic import UUID4
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.permission import Permission
from fief.models.role import Role
from fief.models.user import User


class UserPermission(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "user_permissions"
    __table_args__ = (UniqueConstraint("user_id", "permission_id", "from_role_id"),)

    user_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False
    )
    permission_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Permission.id, ondelete="CASCADE"), nullable=False
    )
    from_role_id: Mapped[UUID4 | None] = mapped_column(
        GUID, ForeignKey(Role.id, ondelete="CASCADE"), nullable=True
    )

    user: Mapped[User] = relationship("User")
    permission: Mapped[Permission] = relationship("Permission")
    from_role: Mapped[Role] = relationship("Role", back_populates="user_permissions")

    def __repr__(self) -> str:
        return f"UserPermission(id={self.id}, user_id={self.user_id}, permission_id={self.permission_id}), from_role_id={self.from_role_id}"



from pydantic import UUID4, AnyHttpUrl

from fief.schemas.generics import BaseModel, CreatedUpdatedAt, UUIDSchema
from fief.schemas.oauth_provider import OAuthProviderEmbedded


class TenantCreate(BaseModel):
    name: str
    registration_allowed: bool = True
    theme_id: UUID4 | None
    logo_url: AnyHttpUrl | None
    oauth_providers: list[UUID4] | None


class TenantUpdate(BaseModel):
    name: str | None
    registration_allowed: bool | None
    theme_id: UUID4 | None
    logo_url: AnyHttpUrl | None
    oauth_providers: list[UUID4] | None


class BaseTenant(UUIDSchema, CreatedUpdatedAt):
    name: str
    default: bool
    slug: str
    registration_allowed: bool
    theme_id: UUID4 | None
    logo_url: AnyHttpUrl | None


class Tenant(BaseTenant):
    oauth_providers: list[OAuthProviderEmbedded]


class TenantEmbedded(BaseTenant):
    pass


class TenantEmailContext(BaseTenant):
    pass



from typing import Generic

from fastapi import Depends

from fief.db import AsyncSession
from fief.dependencies.db import get_main_async_session
from fief.repositories import get_repository
from fief.repositories.base import REPOSITORY


class get_main_repository(Generic[REPOSITORY]):
    def __init__(self, repository_class: type[REPOSITORY]):
        self.repository_class = repository_class

    async def __call__(
        self, session: AsyncSession = Depends(get_main_async_session)
    ) -> REPOSITORY:
        return get_repository(self.repository_class, session)



from datetime import datetime, timezone

from sqlalchemy import select

from fief.models import AuthorizationCode
from fief.repositories.base import BaseRepository, ExpiresAtMixin, UUIDRepositoryMixin


class AuthorizationCodeRepository(
    BaseRepository[AuthorizationCode],
    UUIDRepositoryMixin[AuthorizationCode],
    ExpiresAtMixin[AuthorizationCode],
):
    model = AuthorizationCode

    async def get_valid_by_code(self, code: str) -> AuthorizationCode | None:
        statement = select(AuthorizationCode).where(
            AuthorizationCode.code == code,
            AuthorizationCode.expires_at > datetime.now(timezone.utc),
        )
        return await self.get_one_or_none(statement)

    async def get_by_code(self, code: str) -> AuthorizationCode | None:
        statement = select(AuthorizationCode).where(AuthorizationCode.code == code)
        return await self.get_one_or_none(statement)



"""Add Webhook and WebhookLog

Revision ID: ac62100e37c4
Revises: 30bddbb1abd8
Create Date: 2023-03-02 15:14:05.476824

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "ac62100e37c4"
down_revision = "30bddbb1abd8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_webhooks",
        sa.Column("url", sa.String(length=255), nullable=False),
        sa.Column("secret", sa.String(length=255), nullable=False),
        sa.Column("events", sa.JSON(), nullable=False),
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_webhooks_created_at"),
        "fief_webhooks",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_webhooks_updated_at"),
        "fief_webhooks",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_webhook_logs",
        sa.Column("webhook_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("event", sa.String(length=255), nullable=False),
        sa.Column("attempt", sa.Integer(), nullable=False),
        sa.Column("payload", sa.Text(), nullable=False),
        sa.Column("success", sa.Boolean(), nullable=False),
        sa.Column("response", sa.Text(), nullable=True),
        sa.Column("error_type", sa.String(length=255), nullable=True),
        sa.Column("error_message", sa.Text(), nullable=True),
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["webhook_id"], ["fief_webhooks.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_webhook_logs_created_at"),
        "fief_webhook_logs",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_webhook_logs_updated_at"),
        "fief_webhook_logs",
        ["updated_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_fief_webhook_logs_updated_at"), table_name="fief_webhook_logs"
    )
    op.drop_index(
        op.f("ix_fief_webhook_logs_created_at"), table_name="fief_webhook_logs"
    )
    op.drop_table("fief_webhook_logs")
    op.drop_index(op.f("ix_fief_webhooks_updated_at"), table_name="fief_webhooks")
    op.drop_index(op.f("ix_fief_webhooks_created_at"), table_name="fief_webhooks")
    op.drop_table("fief_webhooks")
    # ### end Alembic commands ###



from wtforms import FieldList, SelectField, StringField, validators

from fief.forms import CSRFBaseForm
from fief.services.oauth_provider import AvailableOAuthProvider


class BaseOAuthProviderForm(CSRFBaseForm):
    provider = SelectField(
        "Provider",
        choices=AvailableOAuthProvider.choices(),
        coerce=AvailableOAuthProvider.coerce,
        validators=[validators.InputRequired()],
    )
    name = StringField("Name")
    openid_configuration_endpoint = StringField(
        "OpenID configuration endpoint",
        validators=[validators.InputRequired(), validators.URL()],
    )
    client_id = StringField("Client ID", validators=[validators.InputRequired()])
    client_secret = StringField(
        "Client Secret", validators=[validators.InputRequired()]
    )
    scopes = FieldList(
        StringField(validators=[validators.InputRequired()]),
        label="Scopes",
        default=[""],
    )

    def __init__(
        self, formdata=None, obj=None, prefix="", data=None, meta=None, **kwargs
    ):
        super().__init__(formdata, obj, prefix, data, meta, **kwargs)
        if self.provider.data != AvailableOAuthProvider.OPENID:
            del self.openid_configuration_endpoint


class OAuthProviderCreateForm(BaseOAuthProviderForm):
    pass


class OAuthProviderUpdateForm(BaseOAuthProviderForm):
    def __init__(
        self, formdata=None, obj=None, prefix="", data=None, meta=None, **kwargs
    ):
        super().__init__(formdata, obj, prefix, data, meta, **kwargs)
        del self.provider



import uuid

import dramatiq
from sqlalchemy.exc import DBAPIError

from fief.logger import logger
from fief.repositories import (
    AuthorizationCodeRepository,
    LoginSessionRepository,
    OAuthSessionRepository,
    RefreshTokenRepository,
    RegistrationSessionRepository,
    SessionTokenRepository,
    WorkspaceRepository,
)
from fief.repositories.base import ExpiresAtRepositoryProtocol
from fief.services.workspace_db import WorkspaceDatabase
from fief.tasks.base import TaskBase

repository_classes: list[type[ExpiresAtRepositoryProtocol]] = [
    AuthorizationCodeRepository,
    LoginSessionRepository,
    OAuthSessionRepository,
    RefreshTokenRepository,
    RegistrationSessionRepository,
    SessionTokenRepository,
]


class CleanupWorkspaceTask(TaskBase):
    __name__ = "cleanup_workspace"

    async def run(self, workspace_id: str):
        latest_revision = WorkspaceDatabase().get_latest_revision()
        workspace = await self._get_workspace(uuid.UUID(workspace_id))
        if workspace.alembic_revision != latest_revision:
            return

        try:
            async with self.get_workspace_session(workspace) as workspace_session:
                for repository_class in repository_classes:
                    repository = repository_class(workspace_session)
                    await repository.delete_expired()
        except ConnectionError:
            logger.warning(
                "Could not connect to workspace", workspace_id=str(workspace.id)
            )
        except DBAPIError as e:
            logger.warning(
                "An error occured while querying workspace",
                workspace_id=str(workspace.id),
                message=str(e),
            )


cleanup_workspace = dramatiq.actor(CleanupWorkspaceTask(), max_retries=0)


class CleanupTask(TaskBase):
    __name__ = "cleanup"

    async def run(self):
        async with self.get_main_session() as session:
            workspace_repository = WorkspaceRepository(session)
            workspaces = await workspace_repository.all()
            for workspace in workspaces:
                self.send_task(cleanup_workspace, workspace_id=str(workspace.id))


cleanup = dramatiq.actor(CleanupTask())



from fastapi import Cookie, Depends

from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.user_field import get_user_create_internal_model, get_user_fields
from fief.dependencies.users import UserManager, get_user_manager
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import RegistrationSession, Tenant, UserField
from fief.repositories import OAuthAccountRepository, RegistrationSessionRepository
from fief.schemas.user import UF, UserCreateInternal
from fief.services.registration_flow import RegistrationFlow
from fief.settings import settings


async def get_registration_flow(
    registration_session_repository: RegistrationSessionRepository = Depends(
        get_workspace_repository(RegistrationSessionRepository)
    ),
    oauth_account_repository: OAuthAccountRepository = Depends(
        get_workspace_repository(OAuthAccountRepository)
    ),
    user_manager: UserManager = Depends(get_user_manager),
    user_create_internal_model: type[UserCreateInternal[UF]] = Depends(
        get_user_create_internal_model
    ),
    user_fields: list[UserField] = Depends(get_user_fields),
) -> RegistrationFlow:
    return RegistrationFlow(
        registration_session_repository,
        oauth_account_repository,
        user_manager,
        user_create_internal_model,
        user_fields,
    )


async def get_optional_registration_session(
    token: str | None = Cookie(None, alias=settings.registration_session_cookie_name),
    registration_session_repository: RegistrationSessionRepository = Depends(
        get_workspace_repository(RegistrationSessionRepository)
    ),
    tenant: Tenant = Depends(get_current_tenant),
) -> RegistrationSession | None:
    if token is None:
        return None

    registration_session = await registration_session_repository.get_by_token(token)
    if registration_session is None:
        return None

    if registration_session.tenant_id != tenant.id:
        return None

    return registration_session



import json

from fastapi import APIRouter, Depends, Query, Request, status
from fastapi.responses import RedirectResponse

from fief.crypto.token import generate_token
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.admin_session import get_admin_session_token
from fief.dependencies.fief import FiefAsyncRelativeEndpoints, get_fief
from fief.dependencies.main_repositories import get_main_repository
from fief.models import AdminSessionToken
from fief.repositories import AdminSessionTokenRepository
from fief.settings import settings

router = APIRouter()


@router.get("/login", name="dashboard.auth:login")
async def login(
    request: Request,
    screen: str = Query("login"),
    fief: FiefAsyncRelativeEndpoints = Depends(get_fief),
):
    url = await fief.auth_url(
        redirect_uri=request.url_for("dashboard.auth:callback"),
        scope=["openid"],
        extras_params={"screen": screen},
    )
    return RedirectResponse(url=url, status_code=status.HTTP_302_FOUND)


@router.get("/callback", name="dashboard.auth:callback")
async def callback(
    request: Request,
    code: str = Query(...),
    fief: FiefAsyncRelativeEndpoints = Depends(get_fief),
    repository: AdminSessionTokenRepository = Depends(
        get_main_repository(AdminSessionTokenRepository)
    ),
):
    tokens, userinfo = await fief.auth_callback(
        code, request.url_for("dashboard.auth:callback")
    )
    token, token_hash = generate_token()
    session_token = AdminSessionToken(
        token=token_hash,
        raw_tokens=json.dumps(tokens),
        raw_userinfo=json.dumps(userinfo),
    )
    await repository.create(session_token)

    response = RedirectResponse(url="/admin/", status_code=status.HTTP_302_FOUND)
    response.set_cookie(
        settings.fief_admin_session_cookie_name,
        token,
        domain=settings.fief_admin_session_cookie_domain,
        secure=settings.fief_admin_session_cookie_secure,
        httponly=True,
    )

    return response


@router.get(
    "/logout",
    name="dashboard.auth:logout",
    dependencies=[Depends(is_authenticated_admin_session)],
)
async def logout(
    request: Request,
    session_token: AdminSessionToken = Depends(get_admin_session_token),
    repository: AdminSessionTokenRepository = Depends(
        get_main_repository(AdminSessionTokenRepository)
    ),
):
    await repository.delete(session_token)

    response = RedirectResponse(
        url=f"//{settings.fief_domain}/logout?redirect_uri={request.base_url}admin/",
        status_code=status.HTTP_302_FOUND,
    )
    response.delete_cookie(
        settings.fief_admin_session_cookie_name,
        domain=settings.fief_admin_session_cookie_domain,
        secure=settings.fief_admin_session_cookie_secure,
        httponly=True,
    )

    return response



from pathlib import Path
from typing import TYPE_CHECKING

from fief.db.workspace import WorkspaceEngineManager, get_workspace_session
from fief.models import EmailTemplate
from fief.repositories.email_template import EmailTemplateRepository
from fief.services.email_template.types import EmailTemplateType

if TYPE_CHECKING:  # pragma: no cover
    from fief.models import Workspace


class EmailTemplateInitializer:
    def __init__(self, repository: EmailTemplateRepository):
        self.repository = repository
        self.templates_dir = Path(__file__).parent / "templates"

    async def init_templates(self):
        if await self.repository.get_by_type(EmailTemplateType.BASE) is None:
            base = EmailTemplate(
                type=EmailTemplateType.BASE,
                subject="",
                content=self._load_template("base.html"),
            )
            await self.repository.create(base)

        if await self.repository.get_by_type(EmailTemplateType.WELCOME) is None:
            welcome = EmailTemplate(
                type=EmailTemplateType.WELCOME,
                subject="Welcome to {{ tenant.name }}",
                content=self._load_template("welcome.html"),
            )
            await self.repository.create(welcome)

        if await self.repository.get_by_type(EmailTemplateType.FORGOT_PASSWORD) is None:
            forgot_password = EmailTemplate(
                type=EmailTemplateType.FORGOT_PASSWORD,
                subject="Reset your {{ tenant.name }}'s password",
                content=self._load_template("forgot_password.html"),
            )
            await self.repository.create(forgot_password)

    def _load_template(self, name: str) -> str:
        with open(self.templates_dir / name) as file:
            return file.read()


async def init_email_templates(workspace: "Workspace"):
    workspace_engine_manager = WorkspaceEngineManager()
    async with get_workspace_session(workspace, workspace_engine_manager) as session:
        email_template_repository = EmailTemplateRepository(session)
        email_template_initializer = EmailTemplateInitializer(email_template_repository)
        await email_template_initializer.init_templates()



import base64
from enum import StrEnum
from typing import TYPE_CHECKING, Any

from httpx_oauth.clients.discord import DiscordOAuth2
from httpx_oauth.clients.facebook import FacebookOAuth2
from httpx_oauth.clients.github import GitHubOAuth2
from httpx_oauth.clients.google import GoogleOAuth2
from httpx_oauth.clients.linkedin import LinkedInOAuth2
from httpx_oauth.clients.microsoft import MicrosoftGraphOAuth2
from httpx_oauth.clients.openid import OpenID
from httpx_oauth.clients.reddit import RedditOAuth2
from httpx_oauth.oauth2 import BaseOAuth2

if TYPE_CHECKING:
    from fief.models import OAuthProvider


class AvailableOAuthProvider(StrEnum):
    DISCORD = "DISCORD"
    FACEBOOK = "FACEBOOK"
    GITHUB = "GITHUB"
    GOOGLE = "GOOGLE"
    LINKEDIN = "LINKEDIN"
    MICROSOFT = "MICROSOFT"
    REDDIT = "REDDIT"
    OPENID = "OPENID"

    def get_display_name(self) -> str:
        if self == AvailableOAuthProvider.OPENID:
            return "OpenID Connect"
        return getattr(OAUTH_PROVIDERS[self], "display_name", "")

    @classmethod
    def choices(cls) -> list[tuple[str, str]]:
        return [(member.value, member.get_display_name()) for member in cls]

    @classmethod
    def coerce(cls, item):
        return cls(str(item)) if not isinstance(item, cls) else item


OAUTH_PROVIDERS: dict[AvailableOAuthProvider, type[BaseOAuth2]] = {
    AvailableOAuthProvider.DISCORD: DiscordOAuth2,
    AvailableOAuthProvider.FACEBOOK: FacebookOAuth2,
    AvailableOAuthProvider.GITHUB: GitHubOAuth2,
    AvailableOAuthProvider.GOOGLE: GoogleOAuth2,
    AvailableOAuthProvider.LINKEDIN: LinkedInOAuth2,
    AvailableOAuthProvider.MICROSOFT: MicrosoftGraphOAuth2,
    AvailableOAuthProvider.REDDIT: RedditOAuth2,
    AvailableOAuthProvider.OPENID: OpenID,
}


def get_oauth_provider_branding(
    oauth_provider: "OAuthProvider",
) -> tuple[str, str | None]:
    provider_class = OAUTH_PROVIDERS[oauth_provider.provider]
    display_name: str = getattr(
        provider_class, "display_name", oauth_provider.name or ""
    )
    logo_svg: str | None = getattr(provider_class, "logo_svg", None)
    if logo_svg:
        logo_svg = base64.b64encode(logo_svg.encode("utf-8")).decode("utf-8")
    return display_name, logo_svg


def get_oauth_provider_service(oauth_provider: "OAuthProvider") -> BaseOAuth2:
    provider = oauth_provider.provider
    oauth_provider_class = OAUTH_PROVIDERS[provider]
    oauth_provider_class_kwargs: dict[str, Any] = {
        "client_id": oauth_provider.client_id,
        "client_secret": oauth_provider.client_secret,
    }
    if provider == AvailableOAuthProvider.OPENID:
        oauth_provider_class_kwargs[
            "openid_configuration_endpoint"
        ] = oauth_provider.openid_configuration_endpoint
    return oauth_provider_class(**oauth_provider_class_kwargs)



import json
import logging
import sys
import uuid
from asyncio import AbstractEventLoop
from collections.abc import Callable
from datetime import timezone
from typing import TYPE_CHECKING, AsyncContextManager, Literal

from loguru import logger
from pydantic import UUID4

from fief.db import AsyncSession
from fief.models import AuditLog, AuditLogMessage, Workspace
from fief.models.generics import M_UUID
from fief.repositories import WorkspaceRepository
from fief.settings import settings

if TYPE_CHECKING:
    from loguru import Logger, Record

LOG_LEVEL = settings.log_level


def stdout_format(record: "Record") -> str:
    record["extra"]["extra_json"] = json.dumps(record["extra"])
    return (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level> - "
        "{extra[extra_json]}"
        "\n{exception}"
    )


class AuditLogger:
    def __init__(
        self,
        logger: "Logger",
        workspace_id: uuid.UUID,
        *,
        admin_user_id: UUID4 | None = None,
        admin_api_key_id: UUID4 | None = None,
    ) -> None:
        self.logger = logger.bind(audit=True, workspace_id=str(workspace_id))
        self.admin_user_id = admin_user_id
        self.admin_api_key_id = admin_api_key_id

    def __call__(
        self,
        message: AuditLogMessage,
        *,
        level="INFO",
        subject_user_id: uuid.UUID | None = None,
        **kwargs,
    ) -> None:
        self.logger.log(
            level,
            message,
            subject_user_id=subject_user_id,
            admin_user_id=self.admin_user_id,
            admin_api_key_id=self.admin_api_key_id,
            **kwargs,
        )

    def log_object_write(
        self,
        operation: Literal[
            AuditLogMessage.OBJECT_CREATED,
            AuditLogMessage.OBJECT_UPDATED,
            AuditLogMessage.OBJECT_DELETED,
        ],
        object: M_UUID,
        subject_user_id: uuid.UUID | None = None,
        **kwargs,
    ) -> None:
        return self(
            operation,
            object_id=str(object.id),
            object_class=type(object).__name__,
            subject_user_id=subject_user_id,
            **kwargs,
        )


class DatabaseAuditLogSink:
    def __init__(
        self,
        get_main_session: Callable[..., AsyncContextManager[AsyncSession]],
        get_workspace_session: Callable[..., AsyncContextManager[AsyncSession]],
    ):
        self.get_main_session = get_main_session
        self.get_workspace_session = get_workspace_session

    async def __call__(self, message):
        record = message.record
        extra = record["extra"]
        workspace_id = extra.get("workspace_id")

        if workspace_id is None:
            return

        workspace = await self._get_workspace(uuid.UUID(workspace_id))

        if workspace is None:
            return

        async with self.get_workspace_session(workspace) as session:
            extra.pop("workspace_id")
            extra.pop("audit")
            subject_user_id = extra.pop("subject_user_id", None)
            object_id = extra.pop("object_id", None)
            object_class = extra.pop("object_class", None)
            admin_user_id = extra.pop("admin_user_id", None)
            admin_api_key_id = extra.pop("admin_api_key_id", None)
            log = AuditLog(
                timestamp=record["time"].astimezone(timezone.utc),
                level=record["level"].name,
                message=record["message"],
                subject_user_id=subject_user_id,
                object_id=object_id,
                object_class=object_class,
                admin_user_id=admin_user_id,
                admin_api_key_id=admin_api_key_id,
                extra=extra,
            )
            session.add(log)
            await session.commit()

    async def _get_workspace(self, workspace_id: uuid.UUID) -> Workspace | None:
        async with self.get_main_session() as session:
            repository = WorkspaceRepository(session)
            return await repository.get_by_id(workspace_id)


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists.
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


logger.remove()
logger.add(
    sys.stdout,
    level=LOG_LEVEL,
    format=stdout_format,
    filter=lambda record: "audit" not in record["extra"],
)

logging.basicConfig(handlers=[InterceptHandler()], level=LOG_LEVEL, force=True)
for uvicorn_logger_name in ["uvicorn", "uvicorn.error"]:
    uvicorn_logger = logging.getLogger(uvicorn_logger_name)
    uvicorn_logger.setLevel(LOG_LEVEL)
    uvicorn_logger.handlers = []
for uvicorn_logger_name in ["uvicorn.access"]:
    uvicorn_logger = logging.getLogger(uvicorn_logger_name)
    uvicorn_logger.setLevel(LOG_LEVEL)
    uvicorn_logger.handlers = [InterceptHandler()]


def init_audit_logger(
    get_main_session: Callable[..., AsyncContextManager[AsyncSession]],
    get_workspace_session: Callable[..., AsyncContextManager[AsyncSession]],
    loop: AbstractEventLoop | None = None,
):
    """
    Initialize the audit logger.

    Needs to be deferred because it relies on a running event loop.
    """
    logger.add(
        DatabaseAuditLogSink(get_main_session, get_workspace_session),
        level=LOG_LEVEL,
        enqueue=True,
        loop=loop,
        filter=lambda r: r["extra"].get("audit") is True,
    )


__all__ = ["init_audit_logger", "logger"]



from pydantic import UUID4
from sqlalchemy import select
from sqlalchemy.sql import Select

from fief.models import Permission, UserPermission
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class PermissionRepository(BaseRepository[Permission], UUIDRepositoryMixin[Permission]):
    model = Permission

    async def get_by_codename(self, codename: str) -> Permission | None:
        statement = select(Permission).where(Permission.codename == codename)
        return await self.get_one_or_none(statement)

    def get_user_permissions_statement(self, user: UUID4) -> Select:
        statement = (
            select(Permission)
            .join(UserPermission, UserPermission.permission_id == Permission.id)
            .where(UserPermission.user_id == user)
        )

        return statement



from sendgrid import SendGridAPIClient, SendGridException
from sendgrid.helpers.mail import Mail

from fief.services.email.base import EmailProvider, SendEmailError, format_address


class Sendgrid(EmailProvider):
    def __init__(self, api_key: str) -> None:
        self._client = SendGridAPIClient(api_key=api_key)

    def send_email(
        self,
        *,
        sender: tuple[str, str | None],
        recipient: tuple[str, str | None],
        subject: str,
        html: str | None = None,
        text: str | None = None,
    ):
        from_email, from_name = sender
        to_email, to_name = recipient
        try:
            message = Mail(
                from_email=format_address(from_email, from_name),
                to_emails=format_address(to_email, to_name),
                subject=subject,
                html_content=html,
                plain_text_content=text,
            )

            self._client.send(message)

        except SendGridException as e:
            raise SendEmailError(str(e)) from e



from datetime import datetime
from typing import TypeVar

from fastapi import Response, status
from fastapi.responses import RedirectResponse
from furl import furl
from pydantic import UUID4

from fief.crypto.access_token import generate_access_token
from fief.crypto.id_token import generate_id_token, get_validation_hash
from fief.crypto.token import generate_token
from fief.dependencies.permission import UserPermissionsGetter
from fief.models import (
    AuthorizationCode,
    Client,
    Grant,
    LoginSession,
    SessionToken,
    Tenant,
    User,
    Workspace,
)
from fief.repositories import (
    AuthorizationCodeRepository,
    GrantRepository,
    LoginSessionRepository,
    SessionTokenRepository,
)
from fief.settings import settings

ResponseType = TypeVar("ResponseType", bound=Response)


class AuthenticationFlow:
    def __init__(
        self,
        authorization_code_repository: AuthorizationCodeRepository,
        login_session_repository: LoginSessionRepository,
        session_token_repository: SessionTokenRepository,
        grant_repository: GrantRepository,
        get_user_permissions: UserPermissionsGetter,
    ) -> None:
        self.authorization_code_repository = authorization_code_repository
        self.login_session_repository = login_session_repository
        self.session_token_repository = session_token_repository
        self.grant_repository = grant_repository
        self.get_user_permissions = get_user_permissions

    async def create_login_session(
        self,
        response: ResponseType,
        *,
        response_type: str,
        response_mode: str,
        redirect_uri: str,
        scope: list[str],
        state: str | None,
        nonce: str | None,
        code_challenge_tuple: tuple[str, str] | None,
        client: Client,
    ) -> ResponseType:
        login_session = LoginSession(
            response_type=response_type,
            response_mode=response_mode,
            redirect_uri=redirect_uri,
            scope=scope,
            state=state,
            nonce=nonce,
            client_id=client.id,
        )
        if code_challenge_tuple is not None:
            code_challenge, code_challenge_method = code_challenge_tuple
            login_session.code_challenge = code_challenge
            login_session.code_challenge_method = code_challenge_method

        login_session = await self.login_session_repository.create(login_session)

        response.set_cookie(
            settings.login_session_cookie_name,
            login_session.token,
            domain=settings.login_session_cookie_domain,
            secure=settings.login_session_cookie_secure,
        )

        return response

    async def delete_login_session(
        self, response: ResponseType, login_session: LoginSession
    ) -> ResponseType:
        response.delete_cookie(
            settings.login_session_cookie_name,
            domain=settings.login_session_cookie_domain,
        )
        await self.login_session_repository.delete(login_session)
        return response

    async def create_session_token(
        self, response: ResponseType, user_id: UUID4
    ) -> ResponseType:
        token, token_hash = generate_token()
        await self.session_token_repository.create(
            SessionToken(token=token_hash, user_id=user_id)
        )
        response.set_cookie(
            settings.session_cookie_name,
            value=token,
            max_age=settings.session_lifetime_seconds,
            domain=settings.session_cookie_domain,
            secure=settings.session_cookie_secure,
            httponly=True,
        )

        return response

    async def rotate_session_token(
        self,
        response: ResponseType,
        user_id: UUID4,
        *,
        session_token: SessionToken | None,
    ) -> ResponseType:
        if session_token is not None:
            await self.session_token_repository.delete(session_token)
        return await self.create_session_token(response, user_id)

    async def create_or_update_grant(
        self, user_id: UUID4, client: Client, scope: list[str]
    ) -> Grant:
        grant = await self.grant_repository.get_by_user_and_client(user_id, client.id)
        if grant is not None:
            grant.scope = scope
            await self.grant_repository.update(grant)
        else:
            grant = await self.grant_repository.create(
                Grant(scope=scope, user_id=user_id, client=client)
            )
        return grant

    async def get_authorization_code_success_redirect(
        self,
        *,
        login_session: LoginSession,
        authenticated_at: datetime,
        user: User,
        client: Client,
        tenant: Tenant,
        workspace: Workspace,
    ) -> RedirectResponse:
        code, code_hash = generate_token()
        c_hash = get_validation_hash(code)
        authorization_code = await self.authorization_code_repository.create(
            AuthorizationCode(
                code=code_hash,
                c_hash=c_hash,
                redirect_uri=login_session.redirect_uri,
                scope=login_session.scope,
                authenticated_at=authenticated_at,
                nonce=login_session.nonce,
                user_id=user.id,
                client_id=client.id,
                expires_at=client.get_authorization_code_expires_at(),
            )
        )

        code_challenge_tuple = login_session.get_code_challenge_tuple()
        if code_challenge_tuple is not None:
            code_challenge, code_challenge_method = code_challenge_tuple
            authorization_code.code_challenge = code_challenge
            authorization_code.code_challenge_method = code_challenge_method

        params = {"code": code}

        if login_session.state is not None:
            params["state"] = login_session.state

        tenant_host = tenant.get_host(workspace.domain)
        access_token: str | None = None

        if login_session.response_type in ["code token", "code id_token token"]:
            permissions = await self.get_user_permissions(user)
            access_token = generate_access_token(
                tenant.get_sign_jwk(),
                tenant_host,
                client,
                user,
                login_session.scope,
                permissions,
                client.access_id_token_lifetime_seconds,
            )
            params["access_token"] = access_token
            params["token_type"] = "bearer"

        if login_session.response_type in ["code id_token", "code id_token token"]:
            id_token = generate_id_token(
                tenant.get_sign_jwk(),
                tenant_host,
                client,
                authenticated_at,
                user,
                client.access_id_token_lifetime_seconds,
                nonce=login_session.nonce,
                c_hash=c_hash,
                access_token=access_token,
                encryption_key=client.get_encrypt_jwk(),
            )
            params["id_token"] = id_token

        parsed_redirect_uri = furl(login_session.redirect_uri)

        if login_session.response_mode == "query":
            parsed_redirect_uri.query.add(params)
        elif login_session.response_mode == "fragment":
            parsed_redirect_uri.fragment.add(args=params)

        return RedirectResponse(
            url=parsed_redirect_uri.url, status_code=status.HTTP_302_FOUND
        )

    @classmethod
    def get_authorization_code_error_redirect(
        cls,
        redirect_uri: str,
        response_mode: str,
        error: str,
        *,
        error_description: str | None = None,
        error_uri: str | None = None,
        state: str | None = None,
    ) -> RedirectResponse:
        parsed_redirect_uri = furl(redirect_uri)

        params = {"error": error}
        if error_description is not None:
            params["error_description"] = error_description
        if error_uri is not None:
            params["error_uri"] = error_uri
        if state is not None:
            params["state"] = state

        if response_mode == "query":
            parsed_redirect_uri.query.add(params)
        elif response_mode == "fragment":
            parsed_redirect_uri.fragment.add(args=params)

        return RedirectResponse(
            url=parsed_redirect_uri.url, status_code=status.HTTP_302_FOUND
        )



from fastapi import APIRouter, Depends

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.email_template import (
    get_email_template_by_id_or_404,
    get_paginated_email_templates,
)
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, EmailTemplate
from fief.repositories.email_template import EmailTemplateRepository
from fief.schemas.generics import PaginatedResults
from fief.services.webhooks.models import EmailTemplateUpdated

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/",
    name="email_templates:list",
    response_model=PaginatedResults[schemas.email_template.EmailTemplate],
)
async def list_email_templates(
    paginated_email_templates: PaginatedObjects[EmailTemplate] = Depends(
        get_paginated_email_templates
    ),
) -> PaginatedResults[schemas.email_template.EmailTemplate]:
    email_templates, count = paginated_email_templates
    return PaginatedResults(
        count=count,
        results=[
            schemas.email_template.EmailTemplate.from_orm(email_template)
            for email_template in email_templates
        ],
    )


@router.get(
    "/{id:uuid}",
    name="email_templates:get",
    response_model=schemas.email_template.EmailTemplate,
)
async def get_email_template(
    email_template: EmailTemplate = Depends(get_email_template_by_id_or_404),
) -> schemas.email_template.EmailTemplate:
    return schemas.email_template.EmailTemplate.from_orm(email_template)


@router.patch(
    "/{id:uuid}",
    name="email_templates:update",
    response_model=schemas.email_template.EmailTemplate,
)
async def update_email_template(
    email_template_update: schemas.email_template.EmailTemplateUpdate,
    email_template: EmailTemplate = Depends(get_email_template_by_id_or_404),
    repository: EmailTemplateRepository = Depends(
        get_workspace_repository(EmailTemplateRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.email_template.EmailTemplate:
    email_template_update_dict = email_template_update.dict(exclude_unset=True)

    for field, value in email_template_update_dict.items():
        setattr(email_template, field, value)

    await repository.update(email_template)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, email_template)
    trigger_webhooks(
        EmailTemplateUpdated,
        email_template,
        schemas.email_template.EmailTemplate,
    )

    return schemas.email_template.EmailTemplate.from_orm(email_template)



"""Add users_count column to Workspace

Revision ID: 3b349d4ffa69
Revises: 4c2bb65882d7
Create Date: 2022-09-29 14:51:46.714954

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "3b349d4ffa69"
down_revision = "4c2bb65882d7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "workspaces",
        sa.Column("users_count", sa.Integer(), server_default="0", nullable=False),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workspaces", "users_count")
    # ### end Alembic commands ###



from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from pydantic import UUID4
from sqlalchemy import select

from fief.crypto.token import get_token_hash
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.main_repositories import get_main_repository
from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.models import AdminAPIKey, Workspace
from fief.repositories import AdminAPIKeyRepository

bearer_scheme = HTTPBearer(auto_error=False)


async def get_optional_admin_api_key(
    authorization: HTTPAuthorizationCredentials | None = Depends(bearer_scheme),
    repository: AdminAPIKeyRepository = Depends(
        get_main_repository(AdminAPIKeyRepository)
    ),
) -> AdminAPIKey | None:
    if authorization is None:
        return None
    token_hash = get_token_hash(authorization.credentials)
    admin_api_key = await repository.get_by_token(token_hash)
    return admin_api_key


async def get_paginated_api_keys(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: AdminAPIKeyRepository = Depends(
        get_main_repository(AdminAPIKeyRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[AdminAPIKey] = Depends(
        get_paginated_objects_getter
    ),
    current_workspace: Workspace = Depends(get_current_workspace),
) -> PaginatedObjects[AdminAPIKey]:
    statement = select(AdminAPIKey).where(
        AdminAPIKey.workspace_id == current_workspace.id
    )
    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_api_key_by_id_or_404(
    id: UUID4,
    repository: AdminAPIKeyRepository = Depends(
        get_main_repository(AdminAPIKeyRepository)
    ),
    current_workspace: Workspace = Depends(get_current_workspace),
) -> AdminAPIKey:
    statement = select(AdminAPIKey).where(
        AdminAPIKey.id == id,
        AdminAPIKey.workspace_id == current_workspace.id,
    )
    api_key = await repository.get_one_or_none(statement)

    if api_key is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return api_key



from fief.schemas import (
    auth,
    client,
    email_template,
    generics,
    oauth_account,
    oauth_provider,
    permission,
    role,
    tenant,
    user,
    user_field,
    user_permission,
    user_role,
    webhook,
    webhook_log,
    well_known,
    workspace,
)

__all__ = [
    "workspace",
    "auth",
    "client",
    "email_template",
    "generics",
    "oauth_account",
    "oauth_provider",
    "permission",
    "role",
    "tenant",
    "user",
    "user_field",
    "user_permission",
    "user_role",
    "webhook",
    "webhook_log",
    "well_known",
]



from datetime import datetime, timezone

from sqlalchemy import select

from fief.models import RegistrationSession
from fief.repositories.base import BaseRepository, ExpiresAtMixin, UUIDRepositoryMixin


class RegistrationSessionRepository(
    BaseRepository[RegistrationSession],
    UUIDRepositoryMixin[RegistrationSession],
    ExpiresAtMixin[RegistrationSession],
):
    model = RegistrationSession

    async def get_by_token(
        self, token: str, *, fresh: bool = True
    ) -> RegistrationSession | None:
        statement = select(RegistrationSession).where(
            RegistrationSession.token == token
        )
        if fresh:
            statement = statement.where(
                RegistrationSession.expires_at > datetime.now(timezone.utc)
            )
        return await self.get_one_or_none(statement)



from enum import StrEnum


class APIErrorCode(StrEnum):
    CANT_DETERMINE_VALID_WORKSPACE = "CANT_DETERMINE_VALID_WORKSPACE"

    WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS = (
        "WORKSPACE_CREATE_MISSING_DATABASE_SETTINGS"
    )
    WORKSPACE_CREATE_INVALID_SSL_MODE = "WORKSPACE_CREATE_INVALID_SSL_MODE"
    WORKSPACE_DB_CONNECTION_ERROR = "WORKSPACE_DB_CONNECTION_ERROR"
    WORKSPACE_DB_OUTDATED_MIGRATION = "WORKSPACE_DB_OUTDATED_MIGRATION"

    CLIENT_CREATE_UNKNOWN_TENANT = "CLIENT_CREATE_UNKNOWN_TENANT"
    CLIENT_HTTPS_REQUIRED_ON_REDIRECT_URIS = "CLIENT_HTTPS_REQUIRED_ON_REDIRECT_URIS"

    OAUTH_PROVIDER_MISSING_OPENID_CONFIGURATION_ENDPOINT = (
        "OAUTH_PROVIDER_MISSING_OPENID_CONFIGURATION_ENDPOINT"
    )
    OAUTH_PROVIDER_REFRESH_TOKEN_NOT_SUPPORTED = (
        "OAUTH_PROVIDER_REFRESH_TOKEN_NOT_SUPPORTED"
    )
    OAUTH_PROVIDER_REFRESH_TOKEN_ERROR = "OAUTH_PROVIDER_REFRESH_TOKEN_ERROR"

    USER_FIELD_SLUG_ALREADY_EXISTS = "USER_FIELD_SLUG_ALREADY_EXISTS"

    USER_CREATE_UNKNOWN_TENANT = "USER_CREATE_UNKNOWN_TENANT"
    USER_CREATE_ALREADY_EXISTS = "USER_CREATE_ALREADY_EXISTS"
    USER_CREATE_INVALID_PASSWORD = "USER_CREATE_INVALID_PASSWORD"

    USER_UPDATE_INVALID_PASSWORD = "USER_UPDATE_INVALID_PASSWORD"
    USER_UPDATE_EMAIL_ALREADY_EXISTS = "USER_UPDATE_EMAIL_ALREADY_EXISTS"

    USER_CREATE_ACCESS_TOKEN_UNKNOWN_CLIENT = "USER_CREATE_ACCESS_TOKEN_UNKNOWN_CLIENT"

    PERMISSION_CREATE_CODENAME_ALREADY_EXISTS = (
        "PERMISSION_CREATE_CODENAME_ALREADY_EXISTS"
    )
    PERMISSION_UPDATE_CODENAME_ALREADY_EXISTS = (
        "PERMISSION_UPDATE_CODENAME_ALREADY_EXISTS"
    )

    ROLE_CREATE_NOT_EXISTING_PERMISSION = "ROLE_CREATE_NOT_EXISTING_PERMISSION"
    ROLE_UPDATE_NOT_EXISTING_PERMISSION = "ROLE_UPDATE_NOT_EXISTING_PERMISSION"

    TENANT_CREATE_NOT_EXISTING_THEME = "TENANT_CREATE_NOT_EXISTING_THEME"
    TENANT_CREATE_NOT_EXISTING_OAUTH_PROVIDER = (
        "TENANT_CREATE_NOT_EXISTING_OAUTH_PROVIDER"
    )
    TENANT_UPDATE_NOT_EXISTING_THEME = "TENANT_UPDATE_NOT_EXISTING_THEME"
    TENANT_UPDATE_NOT_EXISTING_OAUTH_PROVIDER = (
        "TENANT_UPDATE_NOT_EXISTING_OAUTH_PROVIDER"
    )

    USER_PERMISSION_CREATE_NOT_EXISTING_PERMISSION = (
        "USER_PERMISSION_CREATE_NOT_EXISTING_PERMISSION"
    )
    USER_PERMISSION_CREATE_ALREADY_ADDED_PERMISSION = (
        "USER_PERMISSION_CREATE_ALREADY_ADDED_PERMISSION"
    )

    USER_ROLE_CREATE_NOT_EXISTING_ROLE = "USER_ROLE_CREATE_NOT_EXISTING_ROLE"
    USER_ROLE_CREATE_ALREADY_ADDED_ROLE = "USER_ROLE_CREATE_ALREADY_ADDED_ROLE"
